#!/usr/bin/env python
import grpc
import api.v1.spamcheck_pb2 as spam
import api.v1.spamcheck_pb2_grpc as spam_grpc

generic_ham = {
    "text": "Dependency update needed The dependencies for this application are outdated and need to be updated.",
    "user_in_project": False,
    "project": {
        "project_id": 5841855,
        "project_path": "gitlab-com/gl-security/engineering-and-research/automation-team/spam/spamcheck-trigger",
    },
    "user": {
        "emails": [{"email": "integrationtest@example.com", "verified": True}],
        "username": "IntegrationTest",
        "org": "IntegrationTest",
        "id": 123,
        "abuse_metadata": {
            "account_age": 30,
            "spam_score": 0.12,
        },
    },
}
generic_spam = {
    "text": "watch fifa live stream best live streaming [here](https://livestream.com)",
    "user_in_project": False,
    "project": {
        "project_id": 5841855,
        "project_path": "gitlab-com/gl-security/engineering-and-research/automation-team/spam/spamcheck-trigger",
    },
    "user": {
        "emails": [{"email": "integrationtest@example.com", "verified": True}],
        "username": "IntegrationTest",
        "org": "IntegrationTest",
        "id": 23,
        "abuse_metadata": {
            "account_age": 3,
            "spam_score": 0.32,
        },
    },
}

issue_ham = {
    "title": "Dependency update needed",
    "description": "The dependencies for this application are outdated and need to be updated.",
    "user_in_project": False,
    "project": {
        "project_id": 5841855,
        "project_path": "gitlab-com/gl-security/engineering-and-research/automation-team/spam/spamcheck-trigger",
    },
    "user": {
        "emails": [{"email": "integrationtest@example.com", "verified": True}],
        "username": "IntegrationTest",
        "org": "IntegrationTest",
        "id": 23,
        "abuse_metadata": {
            "account_age": 3,
            "spam_score": 0.32,
        },
    },
}
issue_spam = {
    "title": "watch fifa live stream",
    "description": "best live streaming [here](https://livestream.com)",
    "user_in_project": False,
    "project": {
        "project_id": 5841855,
        "project_path": "gitlab-com/gl-security/engineering-and-research/automation-team/spam/spamcheck-trigger",
    },
    "user": {
        "emails": [{"email": "integrationtest@example.com", "verified": True}],
        "username": "IntegrationTest",
        "org": "IntegrationTest",
        "id": 23,
        "abuse_metadata": {
            "account_age": 3,
            "spam_score": 0.32,
        },
    },
}

snippet_ham = {
    "title": " internet time",
    "description": "Swatch Internet Time (or .beat time) is a decimal time concept introduced in 1998 by the Swatch corporation where instead of hours and minutes, the mean solar day is divided into 1000 parts called .beats, with the same time around the globe.",
    "user": {
        "emails": [{"email": "integrationtest@example.com", "verified": True}],
        "username": "IntegrationTest",
        "org": "IntegrationTest",
        "id": 23,
        "abuse_metadata": {
            "account_age": 3,
            "spam_score": 0.32,
        },
    },
    "files": [
        {
            "path": "beats.py",
        }
    ],
}
snippet_spam = {
    "title": "Deep sea fishing Dubai | deep sea fishing",
    "description": "Beach Riders can be easily accessed from a variety of easily accessible tourist destinations and look to provide you the best [deep sea fishing Dubai](https://www.beachridersdubai.com/activities/sport-deep-sea-fishing-in-dubai/) has to offer.[ Deep sea fishing](https://www.beachridersdubai.com/activities/sport-deep-sea-fishing-in-dubai/) is one of the best activities that you can undertake with your loved ones. ",
    "user": {
        "emails": [{"email": "integrationtest@example.com", "verified": True}],
        "username": "IntegrationTest",
        "org": "IntegrationTest",
        "id": 23,
        "abuse_metadata": {
            "account_age": 3,
            "spam_score": 0.32,
        },
    },
    "files": [
        {
            "path": "snippetfile1.txt",
        }
    ],
}


def generic_request(generic):
    generic = spam.Generic(**generic)
    with grpc.insecure_channel("localhost:8001") as channel:
        stub = spam_grpc.SpamcheckServiceStub(channel)
        response = stub.CheckForSpamGeneric(generic)
    return response


def issue_request(issue):
    issue = spam.Issue(**issue)
    with grpc.insecure_channel("localhost:8001") as channel:
        stub = spam_grpc.SpamcheckServiceStub(channel)
        response = stub.CheckForSpamIssue(issue)
    return response


def snippet_request(snippet):
    issue = spam.Snippet(**snippet)
    with grpc.insecure_channel("localhost:8001") as channel:
        stub = spam_grpc.SpamcheckServiceStub(channel)
        response = stub.CheckForSpamSnippet(issue)
    return response


def print_result(response):
    print(f"Verdict: {response.Verdict.Name(response.verdict)}")
    print(f"Score: {response.score}")
    print(f"Reason: {response.reason}")


def main():
    print("Checking ham generic")
    response = generic_request(generic_ham)
    print_result(response)

    print("\nChecking spam generic")
    response = generic_request(generic_spam)
    print_result(response)

    print("\nChecking ham issue")
    response = issue_request(issue_ham)
    print_result(response)

    print("\nChecking spam issue")
    response = issue_request(issue_spam)
    print_result(response)

    print("\nChecking ham snippet")
    response = snippet_request(snippet_ham)
    print_result(response)

    print("\nChecking spam snippet")
    response = snippet_request(snippet_spam)
    print_result(response)


if __name__ == "__main__":
    main()



"""Middleware for gRPC server"""
import abc
import time
from typing import Any, Callable

import grpc
import ulid

from app import logger

log = logger.logger

CORRELATION_ID_KEY = "x-gitlab-correlation-id"
METRIC_API_CALLS = "spamcheck_api_calls"
METRIC_API_FAILURES = "spamcheck_api_failures"

# Metadata fields to include in log messages, if present
_metadata_log_fields = {"user-agent", "retry"}


# pylint: disable=abstract-method
class SpamCheckContext(grpc.ServicerContext):
    """Custom implementation of a context object passed to method implementations."""

    # there is no init in the super class
    # pylint: disable=super-init-not-called
    def __init__(self, context: grpc.ServicerContext):
        self._original_context = context
        self.correlation_id = ulid.new()

    def abort(self, code, details):
        self._original_context.abort(code, details)

    def abort_with_status(self, status):
        self._original_context.abort_with_status(status)

    def add_callback(self, callback):
        self._original_context.add_callback(callback)

    def auth_context(self):
        return self._original_context.auth_context()

    def cancel(self):
        self._original_context.cancel()

    def invocation_metadata(self):
        return self._original_context.invocation_metadata()

    def is_active(self):
        return self._original_context.is_active()

    def peer(self):
        return self._original_context.peer()

    def peer_identities(self):
        return self._original_context.peer_identities()

    def peer_identity_key(self):
        return self._original_context.peer_identity_key()

    def send_initial_metadata(self, initial_metadata):
        self._original_context.send_initial_metadata(initial_metadata)

    def set_code(self, code):
        self._original_context.set_code(code)

    def set_details(self, details):
        self._original_context.set_details(details)

    def set_trailing_metadata(self, trailing_metadata):
        self._original_context.abort(trailing_metadata)

    def time_remaining(self):
        return self._original_context.time_remaining()


# ServerInterceptor code copied from:
# https://github.com/d5h-foss/grpc-interceptor/blob/master/src/grpc_interceptor/server.py#L9
class ServerInterceptor(grpc.ServerInterceptor, metaclass=abc.ABCMeta):
    """Base class for server-side interceptors.
    To implement an interceptor, subclass this class and override the intercept method.
    """

    @abc.abstractmethod
    def intercept(
        self,
        method: Callable,
        request: Any,
        context: grpc.ServicerContext,
        method_name: str,
    ) -> Any:
        """Override this method to implement a custom interceptor.
        You should call method(request, context) to invoke the next handler (either the
        RPC method implementation, or the next interceptor in the list).
        Args:
            method: Either the RPC method implementation, or the next interceptor in
                the chain.
            request: The RPC request, as a protobuf message.
            context: The ServicerContext pass by gRPC to the service.
            method_name: A string of the form "/protobuf.package.Service/Method"
        Returns:
            This should generally return the result of method(request, context), which
            is typically the RPC method response, as a protobuf message. The
            interceptor is free to modify this in some way, however.
        """
        return method(request, context)

    # Implementation of grpc.ServerInterceptor, do not override.
    def intercept_service(self, continuation, handler_call_details):
        """Implementation of grpc.ServerInterceptor.
        This is not part of the grpc_interceptor.ServerInterceptor API, but must have
        a public name. Do not override it, unless you know what you're doing.
        """
        next_handler = continuation(handler_call_details)

        # Only intercept if it's unary:
        if next_handler.request_streaming or next_handler.response_streaming:
            return next_handler

        def invoke_intercept_method(request, context):
            next_interceptor_or_implementation = next_handler.unary_unary
            method_name = handler_call_details.method
            return self.intercept(
                next_interceptor_or_implementation,
                request,
                context,
                method_name,
            )

        return grpc.unary_unary_rpc_method_handler(
            invoke_intercept_method,
            request_deserializer=next_handler.request_deserializer,
            response_serializer=next_handler.response_serializer,
        )


# pylint: disable=too-few-public-methods
class CorrelationIDInterceptor(ServerInterceptor):
    """Add correlation_id to gRPC requests."""

    def intercept(self, method, request, context, method_name):
        ctx = SpamCheckContext(context)
        for item in context.invocation_metadata():
            item = item._asdict()
            if item["key"].lower() == CORRELATION_ID_KEY:
                ctx.correlation_id = item["value"]
                break
        return method(request, ctx)


# pylint: disable=too-few-public-methods
class LoggingInterceptor(ServerInterceptor):
    """Logging interceptor for gRPC requests."""

    def intercept(self, method, request, context, method_name):
        metadata = {
            "correlation_id": context.correlation_id,
            "failed": False,
            "method": method_name,
            "metric": METRIC_API_CALLS,
            "peer": context.peer(),
        }
        for item in context.invocation_metadata():
            item = item._asdict()
            if item["key"] in _metadata_log_fields:
                metadata[item["key"]] = item["value"]

        start_time = time.time()
        try:
            out = method(request, context)
        # Catch and log any unhandled excecptions processing requests
        # pylint: disable=broad-except
        except Exception as ex:
            metadata["failed"] = True
            fields = {
                "metric": METRIC_API_FAILURES,
                "correlation_id": context.correlation_id,
            }
            log.error(str(ex), extra=fields)
            context.set_code(grpc.StatusCode.INTERNAL)
            context.set_details("server error")
            raise
        finally:
            metadata["request_latency_milliseconds"] = (time.time() - start_time) * 1000
            log.info("Processed Request", extra=metadata)
        return out



import unittest
from unittest.mock import patch, PropertyMock

import api.v1.spamcheck_pb2 as spam
from app import config
from app.spammable import generic
from tests.app.helpers import MockContext, MockML


class TestGeneric(unittest.TestCase):
    def test_generic_attrs(self):
        obj = spam.Generic(text="test", type="random_generic")
        g = generic.Generic(obj, MockContext())

        to_dict = g.to_dict()
        self.assertEqual(
            "test", to_dict["description"], "Generic description not set correctly"
        )
        self.assertEqual(
            "random_generic", g.type(), "Generic type not set correctly"
        )

    def test_verdict_project_not_allowed(self):
        g = generic.Generic(spam.Generic(text="test"), MockContext())
        g._project_allowed = False
        v = g.verdict()
        self.assertEqual(
            spam.SpamVerdict.NOOP, v.verdict, "Disallowed project should return NOOP"
        )

    def test_score(self):
        g = generic.Generic(spam.Generic(text="test"), MockContext())

        self.assertEqual(
            spam.SpamVerdict.ALLOW,
            g.calculate_verdict(0.39),
            "Confidence less than 0.4 should be allowed",
        )
        self.assertEqual(
            spam.SpamVerdict.ALLOW,
            g.calculate_verdict(0.41),
            "Confidence between 0.4 and 0,5 should be allowed",
        )
        self.assertEqual(
            spam.SpamVerdict.ALLOW,
            g.calculate_verdict(0.55),
            "Confidence between 0.5 and 0.9 should be allowed",
        )
        self.assertEqual(
            spam.SpamVerdict.ALLOW,
            g.calculate_verdict(0.9),
            "Confidence of 0.9 or greater should be allowed",
        )

    def test_verdict(self):
        generic.classifier = MockML(1.0)
        g = generic.Generic(spam.Generic(text="test"), MockContext())
        g._project_allowed = False
        self.assertEqual(
            spam.SpamVerdict.NOOP,
            g.verdict().verdict,
            "Disallowed project should return NOOP",
        )
        g._project_allowed = True
        g._email_allowed = True
        self.assertEqual(
            spam.SpamVerdict.ALLOW,
            g.verdict().verdict,
            "Allowed email should return ALLOW",
        )
        g._email_allowed = False
        self.assertEqual(
            spam.SpamVerdict.ALLOW,
            g.verdict().verdict,
            "ML inference of 1.0 should be allowed",
        )
        generic.classifier.set_score(0.1)
        self.assertEqual(
            spam.SpamVerdict.ALLOW,
            g.verdict().verdict,
            "ML inference of 0.1 should be allowed",
        )

    def test_verdict_no_ml(self):
        generic.classifier = None
        g = generic.Generic(spam.Generic(text="test"), MockContext())
        g.project_allowed = True
        self.assertEqual(
            spam.SpamVerdict.NOOP,
            g.verdict().verdict,
            "Generic ML not loaded should return NOOP",
        )

    def test_generic_property(self):
        g = generic.Generic(spam.Generic(text="spam"), MockContext())
        g.allowed_domains = {"gitlab.com"}
        self.assertEqual("spam", g.spammable.text, "Generic text should have been set")
        self.assertEqual(False, g._email_allowed, "Blank email should not be allowed")
        args = {"user": {"emails": [{"email": "test@gitlab.com", "verified": True}]}}
        new_generic = spam.Generic(**args)
        g.spammable = new_generic
        self.assertEqual(
            True,
            g._email_allowed,
            "Email should be allowed after updating issue",
        )



"""Store ham and spam for labeling and training."""
import json
import time
from threading import Thread

from google.cloud import storage
from vyper import v

from app import logger

log = logger.logger
bucket = None  # pylint: disable=invalid-name
_BUCKET_KEY = "gcs_bucket"
_stores = {}  # dict containing the DataStore store objects for each spammable type

if v.is_set(_BUCKET_KEY):
    bucket_str = v.get_string(_BUCKET_KEY)
    _storage_client = storage.Client()
    try:
        bucket = _storage_client.get_bucket(bucket_str)
        log.debug("GCS storage enabled")
    except Exception as exp:  # pylint: disable=broad-except
        log.error(f"Failed to initialize GCS bucket: {exp}")
else:
    log.debug("GCS storage disabled")


class DataStore:  # pylint: disable=too-few-public-methods
    """The data store is used to capture ham and spam and write to GCS for labeling and training.
    All issues classified as spam are saved and ham is saved in a balanced manner. Ham storage
    is biased to ensure that spammables that are closest to the SPAM_HAM_THRESHOLD are saved more
    often than other ham.

    Args:
        spammable_type (str): The type of spammable (i.e. issue, snippet).
    """

    MAX_HAM_SIZE = 50
    SPAM_HAM_THRESHOLD = 0.5
    QUESTIONABLE_HAM_BIAS = 3

    def __init__(self, spammable_type: str):
        self.spammable_type = spammable_type.lower()
        self.gcs_file_path = f"spamcheck-unlabeled/{spammable_type}"
        self._questionable_ham = {}
        self._random_ham = {}
        self._questionable_count = 1
        self._labeled_count = 0
        self._ham_spam_count = 0
        self._monitor_objects()

    def save(self, spammable: dict, confidence: float) -> None:
        """Save a spammable for future labeling and training.

        Args:
            spammable (dict): The dictionary representation of a spammable object.
            confidence (float): The confidence value returned from ML inference.
        """
        spammable["type"] = self.spammable_type
        spammable["ml_inference_score"] = confidence

        if confidence < self.SPAM_HAM_THRESHOLD:
            self._add(spammable, confidence)
        else:
            _write_to_gcs(self.gcs_file_path, spammable)
            self._ham_spam_count -= 1

        self._save_ham()

    def _save_ham(self) -> None:
        """Save ham messages until the saved ham is equal to the number of saved spam."""
        while self._ham_spam_count + self._labeled_count < 0:
            ham = self._get()
            if not ham:
                break
            _write_to_gcs(self.gcs_file_path, ham)

            # If _ham_spam_count is less than 0 then we are imbalanced in the unlabeled data we
            # have saved so balance that first.  Otherwise, we are here because of an imbalance
            # in the labeled data so decrement that counter instead.
            if self._ham_spam_count < 0:
                self._ham_spam_count += 1
            else:
                self._labeled_count += 1

    def _add(self, spammable: dict, confidence: float) -> None:
        # First fill up the "questionable ham". This dictionary contains spammables
        # That are closest to the threshold of being classified as spam.
        if len(self._questionable_ham) == 0 or confidence > min(self._questionable_ham):
            self._questionable_ham[confidence] = spammable
            # If the "questionsable ham" is full then remove the lowest confidence spammable
            if len(self._questionable_ham) > self.MAX_HAM_SIZE:
                self._questionable_ham.pop(min(self._questionable_ham))

        # If the "random ham" is not full then add the spammable to that dictionary
        elif len(self._random_ham) < self.MAX_HAM_SIZE:
            self._random_ham[confidence] = spammable

    def _get(self) -> dict:
        ham = None

        # Only get random ham if the modulo of the questionable count is 0.
        # Doing this will create a bias to return the ham messages that are closest
        # to the threshold of marking a spammable as spam versus ham.
        if (
            self._questionable_count % self.QUESTIONABLE_HAM_BIAS == 0
            and len(self._random_ham) > 0
        ):
            self._questionable_count = 1
            # popitem returns a key and value tuple but we have no use for the key
            _, ham = self._random_ham.popitem()

        elif len(self._questionable_ham) > 0:
            self._questionable_count += 1
            ham = self._questionable_ham.pop(max(self._questionable_ham))

        return ham

    def _monitor_objects(self):
        def update_counts():
            prefix = self.gcs_file_path.replace('unlabeled', 'labeled')

            while True:
                count = 0

                log.debug("counting labeled objects for {self.spammable_type}")
                for blob in bucket.list_blobs(prefix=prefix):
                    if '/spam/' in blob.name:
                        count -= 1
                    else:
                        count += 1

                log.debug(f"setting labeled count for {self.spammable_type}: {count}")
                self._labeled_count = count
                time.sleep(900)

        Thread(target=update_counts, daemon=True).start()


def save(spammable_type: str, spammable: dict, confidence: float):
    """Save a spammable for future labeling and training.

    Args:
        spammable_type (str): The type of spammable (i.e. issue, snippet).
        spammable (dict): The dictionary representation of the spammable object.
        confidence (float): The value confidence value returned from ML inference.
    """
    if bucket is None:
        return

    if spammable_type not in _stores:
        _stores[spammable_type] = DataStore(spammable_type)

    func = _stores[spammable_type].save
    thread = Thread(target=func, args=((spammable, confidence)))
    thread.start()


def _write_to_gcs(upload_path: str, data_dict: dict) -> None:
    """Convert a dict to JSON and write to GCS.

    Args:
        upload_path (str): The bucket path to write the blob.
        spammable_dict (dict): The dictionary to save.
    """
    json_str = json.dumps(data_dict)
    filename = data_dict["correlation_id"]
    try:
        blob = bucket.blob(f"{upload_path}/{filename}.json")
        blob.upload_from_string(json_str)
    # pylint: disable=broad-except
    except Exception as ex:
        log.error(f"failed to write to GCS: {ex}")



"""Spamcheck service configuration."""
import argparse
import os

from vyper import v

# pylint: disable=too-many-statements
def load() -> None:
    """Load the configuration options.

    Precedence for config:
        CLI arguments
        Environment variables
        Config file
        Defaults
    """

    if v.get("config_loaded"):
        return

    # Set CLI config options
    parser = argparse.ArgumentParser(description="Application settings")
    parser.add_argument(
        "--env", type=str, choices=["test", "dev", "prod"], help="Application env"
    )
    parser.add_argument("--grpc-addr", type=str, help="Application bind address")
    parser.add_argument(
        "--log-level",
        type=str,
        choices=["debug", "info", "warning", "error", "fatal"],
        help="Application log level",
    )
    parser.add_argument(
        "--ml-classifiers", type=str, help="Directory location for ML classifiers"
    )
    parser.add_argument("--gcs-bucket", type=str, help="Bucket to save spammable data for labeling")
    parser.add_argument("--google-pubsub-project", type=str, help="Google PubSub project")
    parser.add_argument("--google-pubsub-topic", type=str, help="Google PubSub topic")
    parser.add_argument("--tls-certificate", type=str, help="TLS certificate path")
    parser.add_argument("--tls-private-key", type=str, help="TLS private key path")
    parser.add_argument("-c", "--config", type=str, help="Path to config file")

    parser.add_argument("--max-generic-verdict", type=str, help="Maximum verdict for generics")
    parser.add_argument("--max-issue-verdict", type=str, help="Maximum verdict for issues")
    parser.add_argument("--max-snippet-verdict", type=str, help="Maximum verdict for snippets")

    # ignore unknown args. These will be present when running unit tests
    args, _ = parser.parse_known_args()
    v.bind_args(args.__dict__)

    # Set config file options
    v.set_config_type("yaml")
    v.set_config_name("config")
    v.add_config_path("./config")

    # Set environment variable options
    v.set_env_prefix("spamcheck")
    v.automatic_env()

    # Set defaults
    v.set_default("env", "development")
    v.set_default("grpc_addr", "0.0.0.0:8001")
    v.set_default("log_level", "info")
    v.set_default("google_pubsub_topic", "spamcheck")
    v.set_default("tls_certificate", "./ssl/cert.pem")
    v.set_default("tls_private_key", "./ssl/key.pem")
    v.set_default("ml_classifiers", "./classifiers")
    v.set_default("filter.allow_list", {})
    v.set_default("filter.deny_list", {})
    v.set_default("filter.allowed_domains", [])

    v.set_default("max_generic_verdict", "ALLOW")
    v.set_default("max_issue_verdict", "CONDITIONAL_ALLOW")
    v.set_default("max_snippet_verdict", "CONDITIONAL_ALLOW")

    # Load config file
    config_file = None
    if args.config:
        config_file = args.config
    elif v.is_set("config"):
        config_file = v.get_string("config")
    if config_file:
        with open(config_file, encoding="utf-8") as handle:
            config_data = handle.read()
        v.read_config(config_data)
    else:
        try:
            v.read_in_config()
        except FileNotFoundError:
            pass

    if os.path.exists(v.get("tls_private_key")) and os.path.exists(
        v.get("tls_certificate")
    ):
        v.set("tls_enabled", True)

    v.set("config_loaded", True)



"""Process an issue to determine if it is spam or not."""
import api.v1.spamcheck_pb2 as spam

from app import logger, ValidationError
from app.spammable import Spammable


log = logger.logger

# Expecting a module to exist in the directory specified by the ml_classifiers config option.
# i.e {ml_classifiers}/issue/ml
try:
    from issue import classifier
except ModuleNotFoundError as exp:
    log.warning("issue ML classifier not loaded", extra={"error": exp})
    classifier = None  # pylint: disable=invalid-name


def validate(issue: spam.Issue) -> None:
    """Validate that issue contains required fields.

    Raises:
        ValidationError
    """
    if not issue.title:
        raise ValidationError("Issue title is required ")


class Issue(Spammable):
    """Analyze a GitLab issue to determine if it is spam."""

    def __init__(self, issue: spam.Issue, context) -> None:
        validate(issue)
        super().__init__(issue, context, classifier)

Issue.set_max_verdict()



"""Process an snippet to determine if it is spam or not."""
from google.protobuf.json_format import MessageToDict

import api.v1.spamcheck_pb2 as spam

from app import logger
from app.spammable import Spammable
from server.interceptors import SpamCheckContext


log = logger.logger

# Expecting a module to exist in the directory specified by the ml_classifiers config option.
# i.e {ml_classifiers}/snippet/ml
try:
    from issue import classifier
except ModuleNotFoundError as exp:
    log.warning("generic ML classifier not loaded", extra={"error": exp})
    classifier = None  # pylint: disable=invalid-name


class Generic(Spammable):
    """Analyze a generic spammable to determine if it is spam."""

    def __init__(self, spammable: spam.Generic, context: SpamCheckContext) -> None:
        super().__init__(spammable, context, classifier)

    def to_dict(self) -> dict:
        """Return the dictionary representation of the spammable."""
        spammable_dict = MessageToDict(self._spammable)
        spammable_dict["correlation_id"] = str(self.context.correlation_id)
        spammable_dict["title"] = ""
        spammable_dict["description"] = spammable_dict.pop("text")
        return spammable_dict

    def type(self) -> str:
        s_type = self._spammable.type

        if s_type == "":
            s_type = "Generic"

        return s_type

Generic.set_max_verdict()



import unittest
from unittest.mock import patch, PropertyMock

from app import data_store


class TestDataStore(unittest.TestCase):

    @patch.object(data_store.DataStore, "_monitor_objects")
    @patch('app.data_store._write_to_gcs')
    def test_save(self, mock_write, mock_monitor):

        def add_ham():
            # Adding two seen "ham" object to the internal store
            ds._questionable_ham[0.1] = {"a": 1}
            ds._questionable_ham[0.2] = {"b": 2}

        ds = data_store.DataStore(spammable_type="foobar")

        # save a spam object
        ds.save(spammable={}, confidence=ds.SPAM_HAM_THRESHOLD + 0.1)
        self.assertEqual(-1, ds._ham_spam_count, "The spam ham count was not decremented")


        # Save another spam, this should save the two ham objects
        add_ham()
        ds.save(spammable={}, confidence=ds.SPAM_HAM_THRESHOLD + 0.1)
        self.assertEqual(0, ds._ham_spam_count, "The spam ham count should be 0")

        mock_write.assert_any_call(ds.gcs_file_path, {"a": 1})
        mock_write.assert_any_call(ds.gcs_file_path, {"b": 2})

        # Set the counts so there is an imbalance of ham in the labeled dataset
        ds._ham_spam_count = -2
        ds._labeled_count = 10

        add_ham()
        ds.save(spammable={}, confidence=ds.SPAM_HAM_THRESHOLD + 0.1)
        self.assertEqual(-3, ds._ham_spam_count, "We should not have saved any ham")

        # Set the counts so there is an imbalance of spam in the labeled dataset
        ds._ham_spam_count = 1
        ds._labeled_count = -5

        add_ham()
        ds.save(spammable={}, confidence=ds.SPAM_HAM_THRESHOLD + 0.1)
        self.assertEqual(0, ds._ham_spam_count, "Score should have decremented by one when _ham_spam_count is positive and spam was saved")
        self.assertEqual(-3, ds._labeled_count, "We should have saved two ham issues and incremented the _labeled_count")



import unittest
from unittest.mock import patch, PropertyMock

import api.v1.spamcheck_pb2 as spam
from app import config, ValidationError
from app.spammable import issue
from tests.app.helpers import MockContext, MockML

class TestIssue(unittest.TestCase):
    def test_verdict_project_not_allowed(self):
        i = issue.Issue(spam.Issue(title="test"), MockContext())
        i._project_allowed = False
        v = i.verdict()
        self.assertEqual(
            spam.SpamVerdict.NOOP, v.verdict, "Disallowed project should return NOOP"
        )

    def test_score(self):
        s = issue.Issue(spam.Issue(title="test"), MockContext())

        self.assertEqual(
            spam.SpamVerdict.ALLOW,
            s.calculate_verdict(0.39),
            "Confidence less than 0.4 should be allowed",
        )
        self.assertEqual(
            spam.SpamVerdict.CONDITIONAL_ALLOW,
            s.calculate_verdict(0.41),
            "Confidence between 0.4 and 0,5 should be conditionally allowed",
        )
        self.assertEqual(
            spam.SpamVerdict.CONDITIONAL_ALLOW,
            s.calculate_verdict(0.55),
            "Confidence between 0.5 and 0.9 should be disallowed",
        )
        self.assertEqual(
            spam.SpamVerdict.CONDITIONAL_ALLOW,
            s.calculate_verdict(0.9),
            "Confidence of 0.9 or greater should be blocked",
        )

    def test_verdict(self):
        issue.classifier = MockML(1.0)
        i = issue.Issue(spam.Issue(title="test"), MockContext())
        i._project_allowed = False
        self.assertEqual(
            spam.SpamVerdict.NOOP,
            i.verdict().verdict,
            "Disallowed project should return NOOP",
        )
        i._project_allowed = True
        i._email_allowed = True
        self.assertEqual(
            spam.SpamVerdict.ALLOW,
            i.verdict().verdict,
            "Allowed email should return ALLOW",
        )
        i._email_allowed = False
        self.assertEqual(
            spam.SpamVerdict.CONDITIONAL_ALLOW,
            i.verdict().verdict,
            "ML inference of 1.0 should be conditionally allowed",
        )
        issue.classifier.set_score(0.1)
        self.assertEqual(
            spam.SpamVerdict.ALLOW,
            i.verdict().verdict,
            "ML inference of 0.1 should be allowed",
        )

    def test_verdict_no_ml(self):
        issue.classifier = None
        i = issue.Issue(spam.Issue(title="test"), MockContext())
        i.project_allowed = True
        self.assertEqual(
            spam.SpamVerdict.NOOP,
            i.verdict().verdict,
            "Issue ML not loaded should return NOOP",
        )

    def test_issue_property(self):
        i = issue.Issue(spam.Issue(title="spam"), MockContext())
        i.allowed_domains = {"gitlab.com"}
        self.assertEqual("spam", i.spammable.title, "Issue should have been set")
        self.assertEqual(False, i._email_allowed, "Blank email should not be allowed")
        args = {"user": {"emails": [{"email": "test@gitlab.com", "verified": True}]}}
        new_issue = spam.Issue(**args)
        i.spammable = new_issue
        self.assertEqual(
            True,
            i._email_allowed,
            "Email should not be allowed after updating issue",
        )

    def test_validation(self):
        self.assertRaises(ValidationError, issue.Issue, spam.Issue(), MockContext())



"""Process an snippet to determine if it is spam or not."""
import api.v1.spamcheck_pb2 as spam

from app import logger, ValidationError
from app.spammable import Spammable
from server.interceptors import SpamCheckContext


log = logger.logger

# Expecting a module to exist in the directory specified by the ml_classifiers config option.
# i.e {ml_classifiers}/snippet/ml
try:
    from snippet import classifier
except ModuleNotFoundError as exp:
    log.warning("snippet ML classifier not loaded", extra={"error": exp})
    classifier = None  # pylint: disable=invalid-name


def validate(snippet: spam.Snippet) -> None:
    """Validate that snippet contains required fields.

    Raises:
        ValidationError
    """
    if not snippet.title:
        raise ValidationError("Snippet title is required ")


class Snippet(Spammable):
    """Analyze a GitLab snippet to determine if it is spam."""

    def __init__(self, snippet: spam.Snippet, context: SpamCheckContext) -> None:
        validate(snippet)
        super().__init__(snippet, context, classifier)

Snippet.set_max_verdict()



"""Logic to perform spam/ham classification"""
import sys
from typing import Any, List

from google.protobuf.json_format import MessageToDict
from vyper import v

from api.v1.spamcheck_pb2 import SpamVerdict
from app import event, logger, queue, data_store
from server.interceptors import SpamCheckContext

log = logger.logger

classifiers = v.get_string("ml_classifiers")
if classifiers:
    sys.path.append(classifiers)

# pylint: disable=too-few-public-methods
class Spammable:
    """Base class for spammable types."""

    allow_list = v.get("filter.allow_list")
    deny_list = v.get("filter.deny_list")
    allowed_domains = set(v.get("filter.allowed_domains"))

    # Currently maximum allowed value is conditional allow to limit false positives.
    max_verdict = SpamVerdict.CONDITIONAL_ALLOW

    _inference_scores = {
        0.9: SpamVerdict.BLOCK,
        0.5: SpamVerdict.DISALLOW,
        0.4: SpamVerdict.CONDITIONAL_ALLOW,
        0.0: SpamVerdict.ALLOW,
    }

    _verdict_rankings = {
        SpamVerdict.ALLOW: 1,
        SpamVerdict.CONDITIONAL_ALLOW: 2,
        SpamVerdict.DISALLOW: 3,
        SpamVerdict.BLOCK: 4,
    }

    _verdict_mapping = {
        "ALLOW": SpamVerdict.ALLOW,
        "CONDITIONAL_ALLOW": SpamVerdict.CONDITIONAL_ALLOW,
        "DISALLOW": SpamVerdict.DISALLOW,
        "BLOCK": SpamVerdict.BLOCK,
    }

    def __init__(
        self, spammable: Any, context: SpamCheckContext, classifier: None
    ) -> None:
        self.context = context
        self.spammable = spammable
        self.classifier = classifier

    @classmethod
    def set_max_verdict(cls):
        """Set the maximum verdict for the spammable class."""
        key = f"max_{cls.__name__}_verdict".lower()
        verdict_value = v.get_string(key).upper()

        try:
            cls.max_verdict = cls._verdict_mapping[verdict_value]
        except KeyError:
            valid_args = ', '.join(cls._verdict_mapping.keys())
            log.fatal(f"Max verdict must be in [{valid_args}]. Got: {verdict_value}")

    @property
    def spammable(self) -> Any:
        """spam.Spammable: The spammable to analyze for spam"""
        return self._spammable

    @spammable.setter
    def spammable(self, spammable: Any):
        self._spammable = spammable
        self._email_allowed = self.email_allowed(spammable.user.emails)
        if spammable.project:
            self._project_allowed = self.project_allowed(spammable.project.project_id)
        else:
            self._project_allowed = True

    def verdict(self) -> SpamVerdict:
        """Analyze the spammable and determine if spam.

        Returns:
            SpamVerdict
        """

        # If the project is not allowed then this may be an indication that the model
        # does not generalize well to the spammables in that project. In this case we will
        # circumvent evaluating the spammable.
        if not self._project_allowed:
            return self._verdict(SpamVerdict.NOOP, 0.0, "project not allowed", False)
        if not self.classifier:
            return self._verdict(SpamVerdict.NOOP, 0.0, "classifier not loaded", False)

        spammable_dict = self.to_dict()
        confidence = self.classifier.score(spammable_dict)
        data_store.save(self.type(), spammable_dict, confidence)

        if self._email_allowed:
            return self._verdict(SpamVerdict.ALLOW, confidence, "email allowed", True)

        verdict = self.calculate_verdict(confidence)

        return self._verdict(verdict, confidence, "ml inference score", True)

    def calculate_verdict(self, confidence: float) -> SpamVerdict:
        """Convert an ML confidence value to a spam verdict.

        Args:
            confidence (float): The ML confidence value

        Returns:
            SpamVerdict
        """
        for threshold, vdict in self._inference_scores.items():
            if confidence >= threshold:
                return self._maximum_verdict(vdict)

        return SpamVerdict.NOOP

    def _verdict(
        self, verdict: int, confidence: float, reason: str, evaluated: bool
    ) -> SpamVerdict:
        fields = {
            "correlation_id": str(self.context.correlation_id),
            "metric": "spamcheck_verdicts",
            "spammable_type": self.type(),
            "email_allowlisted": self._email_allowed,
            "project_allowed": self._project_allowed,
            "project_path": self._spammable.project.project_path,
            "project_id": self._spammable.project.project_id,
            "user_name": self._spammable.user.username,
            "user_in_project": self._spammable.user_in_project,
            "verdict": SpamVerdict.Verdict.Name(verdict),
            "reason": reason,
            "confidence": confidence,
            "evaluated": evaluated,
        }
        log.info("Verdict calculated", extra=fields)
        if verdict not in (SpamVerdict.ALLOW, SpamVerdict.NOOP):
            evnt = event.Event(event.VERDICT, fields)
            queue.publish(evnt)
        return SpamVerdict(
            verdict=verdict, score=confidence, reason=reason, evaluated=evaluated
        )

    def project_allowed(self, project_id: int) -> bool:
        """Determine if a project should be tested for spam.

        Args:
            project_id (int): The GitLab project ID

        Returns:
            bool
        """
        if len(self.allow_list) != 0:
            if self.allow_list.get(project_id) is not None:
                return True
            return False

        if len(self.deny_list) != 0:
            if self.deny_list.get(project_id) is not None:
                return False
            return True

        return True

    def email_allowed(self, emails: List) -> bool:
        """Determine if a user email should be exempt from spam checking.

        Args:
            emails (list): A list of Emails represented by protobuf objects

        Returns:
            bool
        """
        for email in emails:
            if not "@" in email.email:
                continue
            domain = email.email.split("@")[-1]
            if email.verified and domain in self.allowed_domains:
                return True
        return False

    def type(self) -> str:
        """Get the string representation of the spammable type."""
        return type(self).__name__.lower()

    def to_dict(self) -> dict:
        """Return the dictionary representation of the spammable."""
        spammable_dict = MessageToDict(self._spammable)
        spammable_dict["correlation_id"] = str(self.context.correlation_id)
        return spammable_dict

    def _maximum_verdict(self, verdict: SpamVerdict) -> SpamVerdict:
        max_verdict = self._verdict_rankings[self.max_verdict]
        current_verdict = self._verdict_rankings[verdict]

        if max_verdict < current_verdict:
            return self.max_verdict

        return verdict



import unittest
from unittest.mock import patch, PropertyMock

import api.v1.spamcheck_pb2 as spam
from app import config, ValidationError
from app.spammable import snippet
from tests.app.helpers import MockContext, MockML

class TestSnippet(unittest.TestCase):
    def test_verdict_project_not_allowed(self):
        s = snippet.Snippet(spam.Snippet(title="test"), MockContext())
        s._project_allowed = False
        v = s.verdict()
        self.assertEqual(
            spam.SpamVerdict.NOOP, v.verdict, "Disallowed project should return NOOP"
        )

    def test_score(self):
        s = snippet.Snippet(spam.Snippet(title="test"), MockContext())

        self.assertEqual(
            spam.SpamVerdict.ALLOW,
            s.calculate_verdict(0.39),
            "Confidence less than 0.4 should be allowed",
        )
        self.assertEqual(
            spam.SpamVerdict.CONDITIONAL_ALLOW,
            s.calculate_verdict(0.41),
            "Confidence between 0.4 and 0,5 should be conditionally allowed",
        )
        # Due to valse positives all non allow verdicts were converted to `CONDITIONAL_ALLOW`
        # https://gitlab.com/gitlab-com/gl-security/engineering-and-research/automation-team/spam/spamcheck/-/issues/191
        self.assertEqual(
            spam.SpamVerdict.CONDITIONAL_ALLOW,
            s.calculate_verdict(0.55),
            "Confidence between 0.5 and 0.9 should be conditionally allowed",
        )
        self.assertEqual(
            spam.SpamVerdict.CONDITIONAL_ALLOW,
            s.calculate_verdict(0.9),
            "Confidence of 0.9 or greater should be conditionally allowed",
        )

    def test_verdict(self):
        snippet.classifier = MockML(1.0)
        s = snippet.Snippet(spam.Snippet(title="test"), MockContext())
        s._project_allowed = False
        self.assertEqual(
            spam.SpamVerdict.NOOP,
            s.verdict().verdict,
            "Disallowed project should return NOOP",
        )
        s._project_allowed = True
        s._email_allowed = True
        self.assertEqual(
            spam.SpamVerdict.ALLOW,
            s.verdict().verdict,
            "Allowed email should return ALLOW",
        )
        s._email_allowed = False
        self.assertEqual(
            spam.SpamVerdict.CONDITIONAL_ALLOW,
            s.verdict().verdict,
            "ML inference of 1.0 should be conditionally allowed",
        )
        snippet.classifier.set_score(0.1)
        self.assertEqual(
            spam.SpamVerdict.ALLOW,
            s.verdict().verdict,
            "ML inference of 0.1 should be allowed",
        )

    def test_verdict_no_ml(self):
        snippet.classifier = None
        s = snippet.Snippet(spam.Snippet(title="test"), MockContext())
        s.project_allowed = True
        self.assertEqual(
            spam.SpamVerdict.NOOP,
            s.verdict().verdict,
            "Snippet ML not loaded should return NOOP",
        )

    def test_validation(self):
        self.assertRaises(
            ValidationError, snippet.Snippet, spam.Snippet(), MockContext()
        )



"""Common logging package for spamcheck"""
import logging
import sys
from datetime import datetime
from typing import Any, Optional

from pythonjsonlogger import jsonlogger

from vyper import v


def _log_wrapper(level: str) -> Any:
    def inner(func: Any) -> Any:
        def wrapper(*args, **kwargs) -> Any:
            if kwargs.get("extra") is None:
                kwargs["extra"] = {}

            kwargs["extra"]["time"] = datetime.utcnow().strftime(
                "%Y-%m-%dT%H:%M:%S.%fZ"
            )
            kwargs["extra"]["level"] = level
            kwargs["extra"]["service_name"] = "spamcheck"
            return func(*args, **kwargs)

        return wrapper

    return inner


def _exc_info() -> bool:
    if sys.exc_info()[0]:
        return True
    return False


class Logger:
    """Common JSON logger for spamcheck service."""

    def __init__(self) -> None:
        self.logger = logging.getLogger("spamcheck")
        self.logger.propagate = False
        handler = logging.StreamHandler()
        formatter = jsonlogger.JsonFormatter()
        handler.setFormatter(formatter)
        self.logger.addHandler(handler)
        lvl = v.get_string("log_level").upper()
        level = logging.getLevelName(lvl)
        self.logger.setLevel(level)

    @_log_wrapper("FATAL")
    def fatal(self, msg: str, extra: Optional[dict] = None):
        """Logs a message with level FATAL on the spamcheck logger."""
        self.logger.fatal(msg, exc_info=_exc_info(), extra=extra)
        sys.exit(1)

    @_log_wrapper("ERROR")
    def error(self, msg: str, extra: Optional[dict] = None):
        """Logs a message with level ERROR on the spamcheck logger."""
        self.logger.error(msg, exc_info=_exc_info(), extra=extra)

    @_log_wrapper("WARNING")
    def warning(self, msg: str, extra: Optional[dict] = None):
        """Logs a message with level WARNING on the spamcheck logger."""
        self.logger.warning(msg, extra=extra)

    @_log_wrapper("INFO")
    def info(self, msg: str, extra: Optional[dict] = None):
        """Logs a message with level INFO on the spamcheck logger."""
        self.logger.info(msg, extra=extra)

    @_log_wrapper("DEBUG")
    def debug(self, msg: str, extra: Optional[dict] = None):
        """Logs a message with level DEBUG on the spamcheck logger."""
        self.logger.debug(msg, extra=extra)


logger = Logger()



import unittest
from unittest.mock import patch

import api.v1.spamcheck_pb2 as spam
from vyper import v

from app.spammable import Spammable
from app.spammable.issue import Issue
from tests.app.helpers import MockContext


class TestSpam(unittest.TestCase):
    mock_issue = {
        "title": "mock spam",
        "description": "mock spam",
        "user_in_project": True,
        "project": {
            "project_id": 555,
            "project_path": "mock/spmmable",
        },
        "user": {
            "emails": [{"email": "test@example.com", "verified": True}],
            "username": "test",
            "org": "test",
            "id": 23,
            "abuse_metadata": {
                "account_age": 3,
                "spam_score": 0.32,
            },
        },
    }

    def test_project_allowed(self):
        mock_project = {123: "spamtest/hello"}
        s = Issue(spam.Issue(**self.mock_issue), None)

        allowed = s.project_allowed(555)
        self.assertTrue(allowed, "Empty allow and deny list should return True")

        s.allow_list = mock_project
        allowed = s.project_allowed(555)
        self.assertFalse(allowed, "Project not in allow_list should return False")

        allowed = s.project_allowed(123)
        self.assertTrue(allowed, "Project in allow_list should return True")

        s.allow_list = {}
        s.deny_list = mock_project
        allowed = s.project_allowed(555)
        self.assertTrue(allowed, "Project not in deny_list should return True")

        allowed = s.project_allowed(123)
        self.assertFalse(allowed, "Project in deny_list should return False")

    def test_email_allowed(self):
        def set_emails(email, verified):
            args = {"user": {"emails": [{"email": email, "verified": verified}]}}
            i = spam.Issue(**args)
            return i.user.emails

        s = Issue(spam.Issue(**self.mock_issue), None)
        Spammable.allowed_domains = {"gitlab.com"}
        emails = set_emails("integrationtest@example.com", True)
        allowed = s.email_allowed(emails)
        self.assertFalse(allowed, "Non gitlab email should not be allowed")

        emails = set_emails("integrationtest@gitlab.com", True)
        allowed = s.email_allowed(emails)
        self.assertTrue(allowed, "Verified gitlab email should be allowed")

        emails = set_emails("integrationtest@gitlab.com", False)
        allowed = s.email_allowed(emails)
        self.assertFalse(allowed, "Non-verified gitlab email should not be allowed")

    def test_user_has_id(self):
        s = spam.Issue(**self.mock_issue)
        self.assertEqual(23, s.user.id, "User ID does not match")

    def test_abuse_metadata(self):
        s = spam.Issue(**self.mock_issue)
        error_msg = "Abuse metadata has unepected value"

        self.assertEqual(3, s.user.abuse_metadata["account_age"], error_msg)
        self.assertAlmostEqual(0.32, s.user.abuse_metadata["spam_score"], msg=error_msg)

    def test_to_dict(self):
        spammable = Issue(spam.Issue(**self.mock_issue), MockContext()).to_dict()
        error_msg = "Expected field not seen in spammable dictionary"

        self.assertTrue("user" in spammable, error_msg)
        self.assertTrue("title" in spammable, error_msg)
        self.assertTrue("description" in spammable, error_msg)
        self.assertTrue("project" in spammable, error_msg)
        self.assertTrue("userInProject" in spammable, error_msg)

        user = spammable["user"]

        self.assertTrue("username" in user, error_msg)
        self.assertTrue("id" in user, error_msg)
        self.assertTrue("abuseMetadata" in user, error_msg)
        self.assertTrue("emails" in user, error_msg)
        self.assertTrue("org" in user, error_msg)

    def test_max_verdict(self):
        self.assertEqual(spam.SpamVerdict.CONDITIONAL_ALLOW, Issue.max_verdict, "Unexpected maximum verdict for Issue")

        v.set("max_issue_verdict", "ALLOW")
        Issue.set_max_verdict()
        self.assertEqual(spam.SpamVerdict.ALLOW, Issue.max_verdict, "Unexpected maximum verdict for Issue")



#!/usr/bin/env python
""" Health check CLI for spam service """
import argparse
import os

import grpc

# from api.v1.spamcheck_pb2 import HealthCheckRequest, HealthCheckResponse
# import api.v1.spamcheck_pb2_grpc as spam_grpc
from api.v1.health_pb2 import HealthCheckRequest, HealthCheckResponse
import api.v1.health_pb2_grpc as health_grpc

SPAMCHECK_GRPC_ADDR = os.environ.get("SPAMCHECK_GRPC_ADDR", "0.0.0.0:8001")


def check_liveness():
    """Check to use for Kubernetes liveness probes"""
    with grpc.insecure_channel(SPAMCHECK_GRPC_ADDR) as channel:
        stub = health_grpc.HealthStub(channel)
        response = stub.Check(HealthCheckRequest())

    if not isinstance(response, HealthCheckResponse):
        raise AttributeError(f"Unexpected response type: {type(response)}")

    if response.status == HealthCheckResponse.NOT_SERVING:
        raise RuntimeError("gRPC service not serving")


def check_readiness():
    """Check to use for Kubernetes readiness probes"""


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("--liveness", action="store_true", default=False)
    parser.add_argument("--readiness", action="store_true", default=False)
    args = parser.parse_args()

    if args.liveness:
        check_liveness()
    if args.readiness:
        check_readiness()



"""Event module used to implement cloudevents"""

from cloudevents.http import CloudEvent
from cloudevents.conversion import to_json

VERDICT = "verdict"

class Event(CloudEvent):
    """The Event object which wraps cloudevent functionality.

    Args: data [dict]: The data to build the event from.
    """

    SOURCE = "/spamcheck"
    NAMESPACE = "com.gitlab.spamcheck"

    def __init__(self, _type: str, data: dict):
        attributes = {
            "type": f"{self.NAMESPACE}.{_type}",
            "source": self.SOURCE,
        }
        super().__init__(attributes=attributes, data=data)

    def json(self) -> bytes:
        """Convert Event to a JSON byte array."""
        return to_json(self)



""" gRPC server for spamcheck service. """
from concurrent import futures

import grpc
from grpc_reflection.v1alpha import reflection
from vyper import v

import api.v1.health_pb2 as health
import api.v1.health_pb2_grpc as health_grpc
import api.v1.spamcheck_pb2 as spam
import api.v1.spamcheck_pb2_grpc as spam_grpc
from app import logger, ValidationError
from app.spammable import generic, issue, snippet
from server.interceptors import LoggingInterceptor, CorrelationIDInterceptor

log = logger.logger


# The method names can't be camel case due to generated gRPC code.
#
# pylint: disable=invalid-name
class SpamCheckServicer(spam_grpc.SpamcheckServiceServicer):
    """Handler for gRPC routes."""

    def CheckForSpamGeneric(
        self, request: spam.Generic, context: grpc.ServicerContext
    ) -> spam.SpamVerdict:
        """Route for generic spam."""

        return generic.Generic(request, context).verdict()

    def CheckForSpamIssue(
        self, request: spam.Issue, context: grpc.ServicerContext
    ) -> spam.SpamVerdict:
        """Route for issue spam."""

        try:
            return issue.Issue(request, context).verdict()
        except ValidationError as ex:
            fields = {
                "metric": "spamcheck_validation_errors",
                "correlation_id": context.correlation_id,
            }
            log.warning("Invalid issue", extra=fields)
            context.set_code(grpc.StatusCode.INVALID_ARGUMENT)
            context.set_details(str(ex))
            return spam.SpamVerdict()

    def CheckForSpamSnippet(
        self, request: spam.Snippet, context: grpc.ServicerContext
    ) -> spam.SpamVerdict:
        """Route for snippet spam."""

        try:
            return snippet.Snippet(request, context).verdict()
        except ValidationError as ex:
            fields = {
                "metric": "spamcheck_validation_errors",
                "correlation_id": context.correlation_id,
            }
            log.warning("Invalid snippet", extra=fields)
            context.set_code(grpc.StatusCode.INVALID_ARGUMENT)
            context.set_details(str(ex))
            return spam.SpamVerdict()


class HealthServicer(health_grpc.HealthServicer):
    """Handler for gRPC health check."""

    def Check(self, request: health.HealthCheckRequest, context: grpc.ServicerContext):
        """Urnary health check method"""
        return health.HealthCheckResponse(status=health.HealthCheckResponse.SERVING)

    def Watch(self, request, context):
        """Streaming health check method"""
        yield health.HealthCheckResponse(status=health.HealthCheckResponse.SERVING)


def _fetch_certificate():
    tls_certificate = v.get_string("tls_certificate")
    tls_private_key = v.get_string("tls_private_key")
    with open(tls_private_key, "rb") as file:
        private_key = file.read()
    with open(tls_certificate, "rb") as file:
        certificate_chain = file.read()
    return (private_key, certificate_chain)


def _server(addr: str, tls: bool) -> grpc.Server:
    interceptors = [CorrelationIDInterceptor(), LoggingInterceptor()]
    server = grpc.server(
        futures.ThreadPoolExecutor(max_workers=50), interceptors=interceptors
    )
    if tls:
        creds = grpc.ssl_server_credentials(
            ((_fetch_certificate()),),
            root_certificates=None,
            require_client_auth=False,
        )
        server.add_secure_port(addr, creds)
    else:
        log.warning("TLS certificates not found. Defaulting to insecure channel")
        server.add_insecure_port(addr)

    return server


def serve():
    """Start the gRPC server."""
    env = v.get_string("env")
    tls_enabled = v.get_bool("tls_enabled")
    addr = v.get_string("grpc_addr")
    if addr.isdecimal():
        addr = f"0.0.0.0:{addr}"
    server = _server(addr, tls_enabled)
    spam_grpc.add_SpamcheckServiceServicer_to_server(SpamCheckServicer(), server)
    health_grpc.add_HealthServicer_to_server(HealthServicer(), server)

    # Disable gRPC reflection in production environments
    if env != "production":
        service_names = (
            spam.DESCRIPTOR.services_by_name["SpamcheckService"].full_name,
            reflection.SERVICE_NAME,
        )
        reflection.enable_server_reflection(service_names, server)

    server.start()
    log.info("gRPC server started", extra={"tls_enabled": tls_enabled, "addr": addr})
    try:
        server.wait_for_termination()
    except KeyboardInterrupt:
        log.info("shutting down gRPC server")
        server.stop(None)


if __name__ == "__main__":
    serve()



class MockContext:
    def __init__(self):
        self.correlation_id = "test"


class MockML:
    def __init__(self, score):
        self._score = score

    def score(self, args):
        return self._score

    def set_score(self, score):
        self._score = score



"""Spamcheck queue used to emit spam verdicts"""
from threading import Thread

from google.cloud import pubsub_v1
from google.api_core.exceptions import AlreadyExists
from vyper import v

from app import logger
from app.event import Event

log = logger.logger

queues = set()

class Queue:
    """Base class for Queue implementation. Specific queues should inherit from this class."""
    def enabled(self) -> bool:
        """Determine if a queue is enabled."""
        raise NotImplementedError("Queue not implemented")

    def publish(self, event: Event) -> None:
        """Publish an event to the queue.

        Args:
            event (Event): The Event to publish.
        """
        raise NotImplementedError("Queue not implemented")

class GooglePubSub(Queue):
    """Implementation of the Gooogle PubSub message queue."""

    project_key = "google_pubsub_project"
    topic_key = "google_pubsub_topic"

    def __init__(self):
        self._enabled = False
        if not v.is_set(self.project_key):
            log.debug("Google PubSub disabled")
            return
        project = v.get_string(self.project_key)
        topic = v.get_string(self.topic_key)

        self.publisher = pubsub_v1.PublisherClient()
        self.topic = self.publisher.topic_path(project, topic)
        try:
            self.publisher.create_topic(name=self.topic)
        except AlreadyExists:
            pass
        except Exception as exp: # pylint: disable=broad-except
            log.error(f"Failed to initialize Google PubSub: {exp}")
            return

        log.debug("Google PubSub enabled")
        self._enabled = True

    def enabled(self) -> bool:
        return self._enabled

    def publish(self, event: Event) -> None:
        if self.enabled():
            try:
                future = self.publisher.publish(self.topic, event.json())
                future.result()
            except Exception as exp: # pylint: disable=broad-except
                log.error(f"Google PubSub error: {exp}")

google_pubsub = GooglePubSub()
if google_pubsub.enabled():
    queues.add(google_pubsub)

def publish(event: Event) -> None:
    """Publish an event to all configured queues."""
    for queue in queues:
        thread = Thread(target=queue.publish, args=(event,))
        thread.daemon = True
        thread.start()
